/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 add_images.c tiffread.c tiffwrite.c -lm -ltiff -o add_images

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define VERSION "1.0"

#define FREE_ALL() {				\
    if(NULL != u) free(u);			\
    if(NULL != v) free(v);			\
    if(NULL != out) free(out);			\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\nPointwise addition of the graylevels of two images\n\n");
  printf("Usage: add_images [-f] in1 in2 out\n\n");
  printf("   -f   : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   in1  : first input image (TIFF)\n");
  printf("   in2  : second input kernel (TIFF)\n");
  printf("   out  : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in1 = NULL, *filename_in2 = NULL, *filename_out = NULL, vflag, fflag;
  double *u = NULL,*v = NULL,*out = NULL;
  int i,nx1,ny1,nx2,ny2,err,adr;
  vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (NULL == filename_in1) filename_in1 = argv[i];
    else if (NULL == filename_in2) filename_in2 = argv[i];
    else if (NULL == filename_out) filename_out = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in1) || (NULL == filename_in2) || (NULL == filename_out)) {
    display_usage();
    if (NULL == filename_in1) printf("Error: input 'in1' is missing\n\n");
    else if (NULL == filename_in2) printf("Error: input 'in2' is missing\n\n");
    else printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }


  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx1,&ny1,filename_in1,vflag)) {
    printf("An error occured during the TIFF image reading process (input 'in1').\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if(EXIT_FAILURE == read_tiff_into_double(&v,&nx2,&ny2,filename_in2,vflag)) {
    printf("An error occured during the TIFF image reading process (input 'in2').\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if ((nx1 != nx2) || (ny1 != ny2)) {
    printf("\nError: input images 'in1' and 'in2' must have the same dimensions.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(out = (double*) malloc (nx1*ny1 * sizeof(double)));



  /***********************************/
  /*              KERNEL             */
  /***********************************/
  for(adr=0;adr<nx1*ny1;adr++)
    out[adr] = u[adr]+v[adr];
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/


  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filename_out,out,nx1,ny1,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }


  /* free memory */
  FREE_ALL();
}
