/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 fftconvol.c tiffread.c tiffwrite.c -lfftw3 -lm -ltiff -o fftconvol

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fftw3.h"

#define VERSION "1.0"

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != filter) free(filter);			\
    if(NULL != out) free(out);				\
    if(NULL != dft_u) fftw_free(dft_u);			\
    if(NULL != dft_out) fftw_free(dft_out);		\
    if(NULL != plan_u) fftw_destroy_plan(plan_u);	\
    if(NULL != iplan_out) fftw_destroy_plan(iplan_out);	\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\n2D Fourier-convolution of an image.\n\n");
  printf("Usage: fftconvol [-f] in filter out\n\n");
  printf("   -f     : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   in     : input image (TIFF)\n");
  printf("   filter : convolution filter in the Fourier domain\n");
  printf("   out    : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in = NULL, *filename_filter = NULL, *filename_out = NULL, vflag, fflag;
  double *u = NULL,*filter = NULL,*out = NULL,Omega,val;
  fftw_complex *dft_u = NULL,*dft_out = NULL;
  fftw_plan plan_u = NULL,iplan_out = NULL;
  int i,nx1,ny1,nx,ny,x,y,adr;
  vflag  = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (NULL == filename_in) filename_in = argv[i];
    else if (NULL == filename_filter) filename_filter = argv[i];
    else if (NULL == filename_out) filename_out = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in) || (NULL == filename_filter) || (NULL == filename_out)) {
    display_usage();
    if (NULL == filename_in) printf("Error: input 'in' is missing\n\n");
    else if (NULL == filename_filter) printf("Error: input 'filter' is missing\n\n");
    else printf("Error: output image is missing\n\n");
    return EXIT_FAILURE;
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filename_in,vflag)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if(EXIT_FAILURE == read_tiff_into_double(&filter,&nx1,&ny1,filename_filter,vflag)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if ((nx != nx1) || (ny != ny1)) {
    printf("\nError: 'in' and 'filter' must have the same dimensions.\n\n");
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(out = (double*) malloc (nx*ny * sizeof(double)));
  ASSERT_ALLOC(dft_u = (fftw_complex*) fftw_malloc ((1+nx/2)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_out = (fftw_complex*) fftw_malloc ((1+nx/2)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(plan_u = fftw_plan_dft_r2c_2d(ny,nx,u,dft_u,FFTW_ESTIMATE)); // WARNING: choosing FFTW_MEASURE would destroy u0
  ASSERT_ALLOC(iplan_out = fftw_plan_dft_c2r_2d(ny,nx,dft_out,out,FFTW_ESTIMATE));

  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/

  fftw_execute(plan_u);
  Omega = (double)nx*(double)ny;

  for(x=0;x<=nx/2;x++)
    for(y=0;y<ny;y++) {
      adr = y*(1+nx/2)+x;
      val = filter[y*nx+x];
      dft_out[adr][0] = dft_u[adr][0] * val / Omega;
      dft_out[adr][1] = dft_u[adr][1] * val / Omega;
    }

  fftw_execute(iplan_out);

  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/



  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filename_out,out,nx,ny,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* free memory */
  FREE_ALL();
}
