/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 fftwshift.c tiffread.c tiffwrite.c -lm -ltiff -o fftwshift

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define VERSION "1.0"

#define FREE_ALL() {	 \
  if (NULL != u) { free(u); u = NULL; } \
  if (NULL != v) { free(v); v = NULL; } \
}
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\nInteger periodic shift of an image. Rearranges the FFTW output, moving\n");
  printf("the zero frequency to the center of the spectrum.\n\n");
  printf("Usage: fftwshift [-f] [-i] in out\n\n");
  printf("   -f        : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   -i        : apply invert transformation\n");
  printf("   in        : input image (TIFF)\n");
  printf("   out       : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filenamein = NULL,*filenameout = NULL,fflag = 0,iflag = 0;
  double *u = NULL,*v = NULL;
  int i,err,a,b,nx,ny;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-i") == 0) iflag = 1;
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenameout)) {
    display_usage();
    if(NULL == filenamein) printf("Error: intput 'in' missing\n\n");
    else printf("Error: output 'out' missing\n\n");
    return EXIT_FAILURE;
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filenamein,0)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(v = (double*) malloc (nx*ny * sizeof(double)));


  /*******************************************************/
  /*            KERNEL: shift the input image            */
  /*******************************************************/
  if(iflag) {
    for(b=-ny/2;b<=(ny-1)/2;b++)
      for(a=-nx/2;a<=(nx-1)/2;a++)
	v[((b+ny)%ny)*nx+(a+nx)%nx] = u[(b+ny/2)*nx+(a+nx/2)];
  }
  else {
    for(b=-ny/2;b<=(ny-1)/2;b++)
      for(a=-nx/2;a<=(nx-1)/2;a++)
	v[(b+ny/2)*nx+(a+nx/2)] = u[((b+ny)%ny)*nx+(a+nx)%nx];
  }
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filenameout,v,nx,ny,8,fflag,1,0)) {
    printf("An error occured during the TIFF saving process.\n");
    printf("Free memory and exit program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* free memory */
  FREE_ALL();
}
