/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 perdecomp.c tiffread.c tiffwrite.c -lfftw3 -lm -ltiff -o perdecomp

  References :

  [1] L. Moisan. ``Periodic plus smooth image decomposition''. Journal of Mathematical Imaging and Vision, 2011.
  [2] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [3] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel) recopied from the MegaWave module perdecomp (by Lionel Moisan)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fftw3.h"

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != p) free(p);				\
    if(NULL != s) free(s);				\
    if(NULL != dft_s) fftw_free(dft_s);			\
    if(NULL != plan_s) fftw_destroy_plan(plan_s);	\
    if(NULL != iplan_s) fftw_destroy_plan(iplan_s);	\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

static void display_usage()
{
  printf("\nPeriodic/Smooth image decomposition: in=p+s.\n\n");
  printf("Usage: perdecomp [-f] in p s\n\n");
  printf("   -f  : save outputs as tiff-float instead of unsigned char tiff\n");
  printf("   in  : input image (TIFF)\n");
  printf("   p   : output periodic component (TIFF)\n");
  printf("   s   : output smooth component (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in = NULL, *filename_p = NULL, *filename_s = NULL, vflag, fflag;
  double *u = NULL,*p = NULL,*s = NULL,Omega,a,b,cx,cy;
  fftw_complex *dft_s = NULL;
  fftw_plan plan_s = NULL,iplan_s = NULL;
  int i,nx,ny,x,y,adr;
  vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (NULL == filename_in) filename_in = argv[i];
    else if (NULL == filename_p) filename_p = argv[i];
    else if (NULL == filename_s) filename_s = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in) || (NULL == filename_p) || (NULL == filename_s)) {
    display_usage();
    if (NULL == filename_in) printf("Error: input 'in' is missing\n\n");
    else if (NULL == filename_p) printf("Error: output 'p' is missing\n\n");
    else printf("Error: output 's' is missing\n\n");
    return EXIT_FAILURE;
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(EXIT_FAILURE == read_tiff_into_double(&u,&nx,&ny,filename_in,vflag)) {
    printf("An error occured during the TIFF image reading process.\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  ASSERT_ALLOC(p = (double*) malloc (nx*ny * sizeof(double)));
  ASSERT_ALLOC(s = (double*) calloc (nx*ny , sizeof(double)));

  ASSERT_ALLOC(dft_s = (fftw_complex*) fftw_malloc ((1+nx/2)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(plan_s = fftw_plan_dft_r2c_2d(ny,nx,s,dft_s,FFTW_ESTIMATE));
  ASSERT_ALLOC(iplan_s = fftw_plan_dft_c2r_2d(ny,nx,dft_s,s,FFTW_ESTIMATE));

  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/

  /* fill "boundary image" */
  for(x=0;x<nx;x++) {
    b = u[x]-u[(ny-1)*nx+x]; // = u(x,0) - y(x,ny-1)
    s[x] += b;
    s[(ny-1)*nx+x] += -b;
  }
  for(y=0;y<ny;y++) {
    b = u[y*nx]-u[y*nx+nx-1]; // = u(0,y) - y(nx-1,y)
    s[y*nx] += b;
    s[y*nx+nx-1] += -b;
  }

  /* direct Fourier transform (set dft_s = DFT(s)) */
  fftw_execute(plan_s);

  /* set (DFT(s)(0,0)) = 0 */
  dft_s[0][0] = dft_s[0][1] = 0.;

  /* fill other frequencies */
  cx = 2.*M_PI/(double)nx;
  cy = 2.*M_PI/(double)ny;
  Omega = (double)nx*(double)ny;
  y = 1; // to avoid (0,0) frequency
  for(x=0;x<=nx/2;x++) {
    a = cos(cx*(double)x);
    for(;y<ny;y++) {
      adr = y*(1+nx/2)+x;
      b = 0.5/(2.-a-cos(cy*(double)y));
      dft_s[adr][0] *= b / Omega;
      dft_s[adr][1] *= b / Omega;
    }
    y=0;
  }

  /* inverse Fourier transform (set s = IDFT(dft_s)) */
  fftw_execute(iplan_s); // BEWARE: this command changes the values of dft_s (see FFTW3 documentation)

  /* get p=u-s */
  for(adr=0;adr<nx*ny;adr++)
    p[adr] = u[adr] - s[adr];

  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if (EXIT_FAILURE == save_tiff(filename_p,p,nx,ny,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process (periodic component 'p').\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

    if (EXIT_FAILURE == save_tiff(filename_s,s,nx,ny,8,fflag,1,vflag)) {
    printf("An error occured during the TIFF image saving process (smooth component 's').\n");
    printf("Free memory and exit the program.\n\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }


  /* free memory */
  FREE_ALL();
}
