/**************************************************************************
   STV-TOOLS: Shannon Total Variation based image processing
   Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation (with gcc): gcc -O3 stvdeconv.c tiffread.c tiffwrite.c shannon_tools.c convol_tools.c -lfftw3 -lm -ltiff -o stvdeconv

  References :

  [1] R. Abergel and L. Moisan, ``The Shannon Total Variation'', preprint MAP5 n° 2016-19, July 2016.
  [2] L. Moisan. How to discretize the total variation of an image? Proceedings in Applied Mathematics and Mechanics, 2007.

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fftw3.h"

#define VERSION "1.0"

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != gx) free(gx);				\
    if(NULL != gy) free(gy);				\
    if(NULL != px) free(px);				\
    if(NULL != py) free(py);				\
    if(NULL != u0) free(u0);				\
    if(NULL != ker) free(ker);				\
    if(NULL != Asq) free(Asq);				\
    if(NULL != ubar) free(ubar);			\
    if(NULL != Aubar) free(Aubar);			\
    if(NULL != dft_ubar) fftw_free(dft_ubar);		\
    if(NULL != dft_gx) fftw_free(dft_gx);		\
    if(NULL != dft_gy) fftw_free(dft_gy);		\
    if(NULL != dft_px) fftw_free(dft_px);		\
    if(NULL != dft_py) fftw_free(dft_py);		\
    if(NULL != dft_div) fftw_free(dft_div);		\
    if(NULL != plan_u) fftw_destroy_plan(plan_u);	\
    if(NULL != plan_ubar) fftw_destroy_plan(plan_ubar); \
    if(NULL != plan_px) fftw_destroy_plan(plan_px);	\
    if(NULL != plan_py) fftw_destroy_plan(plan_py);	\
    if(NULL != iplan_gx) fftw_destroy_plan(iplan_gx);	\
    if(NULL != iplan_gy) fftw_destroy_plan(iplan_gy);	\
    if(NULL != iplan_div) fftw_destroy_plan(iplan_div); \
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }
#define MIN(a,b)          (((a)>(b)) ? (b) : (a))
#define MAX(a,b)          (((a)>(b)) ? (a) : (b))


/* eval_energy : compute E(u) = ||Au-u0||_2^2 + lambda*STVn(u) */
static double eval_energy(Au,u0,gx,gy,nx0,ny0,lambda,n,Omega_n)
     double *Au,*u0,*gx,*gy,lambda;
     int nx0,ny0,n,Omega_n;
{
  int adr;
  double E1=0.,E2=0.,delta;

  for(adr=0;adr<nx0*ny0;adr++) {
    delta = Au[adr]-u0[adr];
    E1 += delta*delta;
  }

  for(adr=0;adr<Omega_n;adr++) E2 += hypot(gx[adr],gy[adr]);

  return E1 + lambda/((double)n*(double)n)*E2;
}

static void update_primal(u,ubar,div,Asq,lambda,n,tau,Omega)
     double *u,*ubar,*div,*Asq,lambda,tau,Omega;
     int n;
{
  double cof;
  int adr;

  cof = tau*lambda/((double)n*(double)n);

  for(adr=0;adr<Omega;adr++) {
    ubar[adr] = -u[adr];
    u[adr] += cof*div[adr] - tau*Asq[adr];
    ubar[adr] += 2.*u[adr];
  }
}


static void update_dual_p(px,py,gx,gy,lambda,n,sigma,Omega_n)
     double *px,*py,*gx,*gy,lambda,sigma,Omega_n;
     int n;
{
  int adr;
  double nrm,cof = sigma*lambda/((double)n*(double)n);

  for(adr=0;adr<Omega_n;adr++) {
    px[adr] += cof*gx[adr];
    py[adr] += cof*gy[adr];
    nrm = px[adr]*px[adr] + py[adr]*py[adr];
    if(nrm > 1.) {
      nrm = sqrt(nrm);
      px[adr] /= nrm;
      py[adr] /= nrm;
    }
  }
}


static void update_dual_q(q,Aubar,u0,sigma,nx0,ny0)
     double *q,*Aubar,*u0,sigma;
     int nx0,ny0;
{
  int adr;
  double cof1,cof2;

  cof1 = 2./(2.+sigma);
  cof2 = sigma * cof1;

  for(adr=0;adr<nx0*ny0;adr++)
    q[adr] = cof1*q[adr] + cof2*(Aubar[adr]-u0[adr]);
}


static void display_usage()
{
  printf("\nSTV-based image deconvolution.\n\n");
  printf("Usage: stvdeconv [-N Niter] [-W lambda] [-n oversampling] [-t tau] [-s sigma] [-p] [-f] [-v] [-vv] in ker out\n\n");
  printf("   -N Niter        : (default = 100) number of iterations\n");
  printf("   -W lambda       : (default = 10.) set regularity parameter (TV weight)\n");
  printf("   -n oversampling : (default = 2) oversampling factor (use the STV_n regularizer)\n");
  printf("   -t tau          : primal step for the Chambolle-Pock Algorithm\n");
  printf("   -s sigma        : dual step for the Chambolle-Pock Algorithm\n");
  printf("   -p              : preserve from output cropping (output size will be bigger than input size)\n");
  printf("   -f              : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   -v              : light verbose mode (display progression)\n");
  printf("   -vv             : heavy verbose mode (compute & display energy evolution)\n");
  printf("   in              : blurry and noisy input image (TIFF)\n");
  printf("   ker             : blur kernel (TIFF)\n");
  printf("   out             : output image (TIFF)\n\n");
}


int main(int argc, char **argv)
{
  char *filename_in, *filename_ker, *filename_out, *Wvalue, *Nvalue, *nvalue, *tvalue, *svalue, vflag, fflag, vvflag, pflag;
  double *u0,*u,*ubar,*px,*py,*gx,*gy,*div,*ker,*q,*Aubar,*Asq,lambda,sigma,tau,L,Omega,Omega_n,n1k,E;
  fftw_complex *dft_div,*dft_ubar,*dft_px,*dft_py,*dft_gx,*dft_gy;
  fftw_plan plan_u,plan_ubar,plan_px,plan_py,iplan_gx,iplan_gy,iplan_div;
  int n,niter,i,err,nx0,ny0,kx,ky,dx1,dx2,dy1,dy2,nx,ny,Nx,Ny,half_nx,half_ny,half_Nx,half_Ny,x,y,adr;

  vvflag = vflag = fflag = pflag = 0;
  filename_in = filename_ker = filename_out = Wvalue = Nvalue = nvalue = tvalue = svalue = NULL;
  u0 = u = ubar = px = py = gx = gy = div = ker = q = Aubar = Asq = NULL;
  dft_div = dft_ubar = dft_px = dft_py = dft_gx = dft_gy = NULL;
  plan_u = plan_ubar = plan_px = plan_py = iplan_gx = iplan_gy = iplan_div = NULL;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-v") == 0) vflag = 1;
    else if (strcmp(argv[i],"-vv") == 0) vvflag = 1;
    else if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-p") == 0) pflag = 1;
    else if (strcmp(argv[i],"-N") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -N requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Nvalue = argv[i+1]; err = sscanf(Nvalue,"%d",&niter); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -N\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-W") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -W requires an argument\n"); return EXIT_FAILURE; }
      else {
	Wvalue = argv[i+1]; err = sscanf(Wvalue,"%lf",&lambda); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -W\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-n") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -n requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	nvalue = argv[i+1]; err = sscanf(nvalue,"%d",&n); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -n\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-t") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -t requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	tvalue = argv[i+1]; err = sscanf(tvalue,"%lf",&tau); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -t\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-s") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -s requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	svalue = argv[i+1]; err = sscanf(svalue,"%lf",&sigma); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -s\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (NULL == filename_in) filename_in = argv[i];
    else if (NULL == filename_ker) filename_ker = argv[i];
    else if (NULL == filename_out) filename_out = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* check consistency */
  if ((NULL == filename_in) || (NULL == filename_ker) || (NULL == filename_out)) {
    display_usage();
    if (NULL == filename_in) printf("Error: input 'in' is missing\n\n");
    else if(NULL == filename_ker) printf("Error: input 'ker' is missing\n\n");
    else if(NULL == filename_out) printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(vflag || vvflag) {
    printf("Load input image and allocate memory\n");
    printf("====================================\n");
  }

  if(EXIT_FAILURE == read_tiff_into_double(&u0,&nx0,&ny0,filename_in,vflag || vvflag)) {
    if(vflag || vvflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  if(EXIT_FAILURE == read_tiff_into_double(&ker,&kx,&ky,filename_ker,vflag || vvflag)) {
    if(vflag || vvflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* set default values */
  if(NULL == Nvalue) niter = 100;
  if(NULL == Wvalue) lambda = 10;
  if(NULL == nvalue) n = 2;
  if((NULL == svalue) || (NULL == tvalue)) {
    for(n1k=0,i=0;i<kx*ky;i++) n1k += fabs(ker[i]);
    L = hypot(lambda*M_PI*sqrt(2)/(double)n,n1k);
    if(NULL == tvalue) { tau = 0.99/L; }
    if(NULL == svalue) { sigma = 0.99/L; }
  }

  /* deal with verbose mode */
  if(vflag || vvflag) {
    printf("\nInput parameters\n");
    printf("================\n");
    printf("number of iterations: Niter = %d\n",niter);
    printf("regularity parameter: lambda = %.17g\n",lambda);
    printf("oversampling factor: n = %d\n",n);
    printf("primal step for Chambolle-Pock Algorithm: tau = %.17g\n",tau);
    printf("dual step for Chambolle-Pock Algorithm: sigma = %.17g\n\n",sigma);
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  dx1 = kx/2; dx2 = (kx-1)/2;
  dy1 = ky/2; dy2 = (ky-1)/2;
  nx = nx0+dx1+dx2; ny = ny0+dy1+dy2;
  Nx = n*nx; Ny = n*ny; Omega = nx*ny; Omega_n = Nx*Ny;
  half_nx = nx/2; half_ny = ny/2; half_Nx = Nx/2; half_Ny = Ny/2;

  if(vflag || vvflag) { printf("memory allocation: in progress"); fflush(stdout);}

  ASSERT_ALLOC(u     = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(ubar  = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(div   = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(Asq   = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(gx    = (double*) malloc(Omega_n * sizeof(double)));
  ASSERT_ALLOC(gy    = (double*) malloc(Omega_n * sizeof(double)));
  ASSERT_ALLOC(px    = (double*) calloc(Omega_n , sizeof(double)));
  ASSERT_ALLOC(py    = (double*) calloc(Omega_n , sizeof(double)));
  ASSERT_ALLOC(q     = (double*) calloc(nx0*ny0 , sizeof(double)));
  ASSERT_ALLOC(Aubar = (double*) calloc(nx0*ny0 , sizeof(double)));

  ASSERT_ALLOC(dft_gx   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_gy   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_px   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_py   = (fftw_complex*) fftw_malloc((1+half_Nx)*Ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_ubar = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));
  ASSERT_ALLOC(dft_div  = (fftw_complex*) fftw_malloc((1+half_nx)*ny * sizeof(fftw_complex)));

  ASSERT_ALLOC(plan_u    = fftw_plan_dft_r2c_2d(ny,nx,u,dft_ubar,FFTW_MEASURE)); // set dft_ubar = DFT(u) (not DFT(ubar))
  ASSERT_ALLOC(plan_ubar = fftw_plan_dft_r2c_2d(ny,nx,ubar,dft_ubar,FFTW_MEASURE));
  ASSERT_ALLOC(plan_px   = fftw_plan_dft_r2c_2d(Ny,Nx,px,dft_px,FFTW_MEASURE));
  ASSERT_ALLOC(plan_py   = fftw_plan_dft_r2c_2d(Ny,Nx,py,dft_py,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_div = fftw_plan_dft_c2r_2d(ny,nx,dft_div,div,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_gx  = fftw_plan_dft_c2r_2d(Ny,Nx,dft_gx,gx,FFTW_MEASURE));
  ASSERT_ALLOC(iplan_gy  = fftw_plan_dft_c2r_2d(Ny,Nx,dft_gy,gy,FFTW_MEASURE));

  if(vflag || vvflag) { printf("\33[2K\rmemory allocation: done\n\n"); fflush(stdout); }




  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/
  if(vflag || vvflag) {
    printf("Start the image restoration process\n");
    printf("===================================\n");
  }

  /* initialize u & ubar */
  for(adr=y=0;y<ny;y++)
    for(x=0;x<nx;x++,adr++) {
      u[adr] = ubar[adr] = u0[(MAX(dy2,MIN(ny-1-dy1,y))-dy2)*nx0+(MAX(dx2,MIN(nx-1-dx1,x))-dx2)];
    }

  /**** main loop ****/
  for(i=1;i<=niter;i++) {

    /* compute Shannon gradient of ubar into gx and gy */
    compute_sgrad(dft_ubar,dft_gx,dft_gy,plan_ubar,iplan_gx,iplan_gy,nx,ny,Nx,Ny);
    /* update dual variable p = (px,py) and q */
    conv(Aubar,ubar,ker,nx,ny,kx,ky);
    update_dual_p(px,py,gx,gy,lambda,n,sigma,Omega_n);
    update_dual_q(q,Aubar,u0,sigma,nx0,ny0);
    adjoint_conv(Asq,q,ker,nx,ny,kx,ky);
    /* compute Shannon divergence into div */
    compute_sdiv(dft_div,dft_px,dft_py,plan_px,plan_py,iplan_div,nx,ny,Nx,Ny);
    /* update primal variables u and ubar */
    update_primal(u,ubar,div,Asq,lambda,n,tau,Omega);

    if(vflag) {
      printf("\33[2K\riteration %d/%d",i,niter);
      fflush(stdout);
    }
    else if(vvflag) {
      compute_sgrad(dft_ubar,dft_gx,dft_gy,plan_u,iplan_gx,iplan_gy,nx,ny,Nx,Ny); // set (gx,gy) = Grad_n(u) (not ubar)
      conv(Aubar,u,ker,nx,ny,kx,ky); // set Aubar = convolution between ker and u (not ubar)
      E = eval_energy(Aubar,u0,gx,gy,nx0,ny0,lambda,n,(int)Omega_n);
      printf("iteration %d/%d : E = %.10g\n",i,niter,E);
      fflush(stdout);
    }
  }

  if(vflag || vvflag) printf("\n\n");
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if(vflag || vvflag) {
    printf("Save restored image as a TIFF image and free memory\n");
    printf("===================================================\n");
  }

  if(pflag) { // output = u
    if (EXIT_FAILURE == save_tiff(filename_out,u,nx,ny,8,fflag,1)) {
      if (vflag || vvflag) printf("free memory and exit the program.\n\n");
      else printf("An error occured (try verbose mode using option -v to get more informations)\n");
      FREE_ALL();
      return EXIT_FAILURE;
    }
  }
  else {
    // set firsts nx0*ny0 values of ubar equal to the restriction of u to the domain of u0
    for(x=dx2;x<nx-dx1;x++)
      for(y=dy2;y<ny-dy2;y++)
	ubar[(y-dy2)*nx0+(x-dx2)] = u[y*nx+x];
    // save nx0*ny0 firsts graylevels as output
    if (EXIT_FAILURE == save_tiff(filename_out,ubar,nx0,ny0,8,fflag,1)) {
      if (vflag || vvflag) printf("free memory and exit the program.\n\n");
      else printf("An error occured (try verbose mode using option -v to get more informations)\n");
      FREE_ALL();
      return EXIT_FAILURE;
    }
  }

  /* free memory */
  if(vflag || vvflag) {
    printf("free memory: in progress");
    fflush(stdout);
  }

  FREE_ALL();

  if(vflag || vvflag) {
    printf("\33[2K\rfree memory: done\n\n");
    fflush(stdout);
  }
}
